@@ -61,8 +61,7 @@ gst_gio_base_src_base_init (gpointer gclass)
   GST_DEBUG_CATEGORY_INIT (gst_gio_base_src_debug, "gio_base_src", 0,
       "GIO base source");
 
-  gst_element_class_add_pad_template (element_class,
-      gst_static_pad_template_get (&src_factory));
+  gst_element_class_add_static_pad_template (element_class, &src_factory);
 }
 
 static void
